--[[
    Polling for interrupts
]]

local ADDON_NAME, ns = ...

local Printer = LibStub:GetLibrary("OakPrint-1.0")(ADDON_NAME)
ns.Printer = Printer

local Data = ns.Data

--------------------------------------------------------------------------------
-- Dataobject
--------------------------------------------------------------------------------

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local dataobj = ldb:NewDataObject("OakInterrupts",
        {
            type = "OakInterrupt",
            polling = false
        })
ns.dataobj = dataobj

--------------------------------------------------------------------------------
-- Helper functions/tables
--------------------------------------------------------------------------------

local specTable = {}
ns.specTable = specTable

local classTable = {}
ns.classTable = classTable

local nameTable = {}
ns.nameTable = nameTable

do
    local function IsPlayerGUID(str)
        return str and string.match(str, 'Player-.*')
    end

    setmetatable(nameTable, {
        __index = function(t, k)
            if k ~= nil then
                name = IsPlayerGUID(k) and select(6, GetPlayerInfoByGUID(k)) or nil
                t[k] = name
                return name
            end
        end})
end

local _interruptTable = {}
local interruptTable = {}
ns.interruptTable = interruptTable

setmetatable(interruptTable, {
    __index = _interruptTable,
    __newindex = function(t, k, v)
        assert(not (k == nil), "interruptTable: tried to set nil key")
        local name = nameTable[k]
        if v == nil then -- unregister
            _interruptTable[k] = nil
            dataobj[name] = nil
            return
        end

        local class = classTable[k]
        local spellID = Data.spellOfClass[class] or Data.spellOfSpec[specTable[k]]
        local spell = Data.spellTable[spellID]

        if spell then
            local cd = spell.cooldown
            local time = v > 0 and v + cd or _interruptTable[k] or 0
            Printer.debug(2,
                       "Setting interrupt for %s with expiration time %d",
                       name, time)
            _interruptTable[k] = time
            dataobj[name] = time -- set exported data
        end
    end,
    __call = function()
        return pairs(_interruptTable)
    end
})


--------------------------------------------------------------------------------
-- Controller
--------------------------------------------------------------------------------

local OIB = CreateFrame('Frame')

function OIB:UNIT_SPELLCAST_SUCCEEDED(unitID, spell, _, lineID, spellID)
    local guid = UnitGUID(unitID)
    if type(interruptTable[guid]) == "number" and Data.spellTable[spellID] then
        Printer.debug(5, "Tracked spell (%s) cast by %s", spell, nameTable[guid])
        interruptTable[guid] = GetTime()
    end
end

function OIB.Enable()
    OIB:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    ns.Inspect.Enable()
    ns.Inspect.PollGroup()
end

function OIB.Disable()
    OIB:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    ns.Inspect.Disable()
    local names = {}
    for n, _ in interruptTable() do
        if n ~= "type" and n ~= "polling" and n ~= "filter_func" then
            table.insert(names, n)
        end
    end
    for _, n in ipairs(names) do
        interruptTable[n] = nil
    end
end

function OIB:ADDON_LOADED(addon_name)
    if addon_name == ADDON_NAME then
        OIB.Enable()
        OIB:UnregisterEvent('ADDON_LOADED')
    end
end

OIB:RegisterEvent('ADDON_LOADED')

OIB:SetScript('OnEvent', function(self, event, ...) self[event](self, ...) end)

--------------------------------------------------------------------------------
-- Slash commands
--------------------------------------------------------------------------------

local function Dump(heading, t)
    if t == dataobj then
        _, t = ldb:pairs(t)
    end
    Printer.print("--- " .. heading .. " ---")
    for k, v in next, t do
       Printer.print("%s: %s", k, tostring(v))
    end
end

local misc_table = {}
local function UpdateMiscTable()
    misc_table["Update script"] = ns.Inspect.state.script
    misc_table["Inspection target"] = nameTable[ns.Inspect.state.current] or 'none'
end

local commands = LibStub:GetLibrary("OakSlash-1.0")(ADDON_NAME)
local t = {
    enable = OIB.Enable,
    disable = OIB.Disable,
    poll = function() ns.Inspect.PollGroup(); ns.Inspect.PerformInspections() end,
    filter = { ranged = function(...) ns.Inspect.filter(true, 'ranged') end,
               none = function() ns.Inspect.filter('reset') end},
    test = function() ns.Inspect.TestPoll() end,
    dump = { all = function()
                       UpdateMiscTable()
                       Dump("Miscellaneous", misc_table)
                       Dump("Exported Data", dataobj)
                       Dump("Notify Register", ns.Inspect.state.register)
                   end,
             data = function() Dump("Exported Data", dataobj) end,
             inspect = function()
                           Dump("Notify Register", ns.Inspect.state.register)
                       end,
             misc = function()
                        UpdateMiscTable()
                        Dump("Miscellaneous", misc_table)
                    end},
    debug = Printer.SetDebug
}
commands:NewCommand("oib", t, "oakinterruptsbroker")
